@@ -186,7 +186,7 @@ class ManholeTestCase(unittest.TestCase):
                     proc.reset()
                     self.assertManholeRunning(proc, uds_path)
 
-    def assertManholeRunning(self, proc, uds_path, oneshot=False):
+    def assertManholeRunning(self, proc, uds_path, oneshot=False, extra=None):
         sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
         sock.settimeout(0.5)
         sock.connect(uds_path)
@@ -204,6 +204,8 @@ class ManholeTestCase(unittest.TestCase):
                     self._wait_for_strings(proc.read, 10,
                         'from PID:%s UID:%s' % (os.getpid(), os.getuid()),
                     )
+                    if extra:
+                        extra(sock)
                     sock.shutdown(socket.SHUT_RDWR)
         finally:
             sock.close()
@@ -359,27 +361,70 @@ class ManholeTestCase(unittest.TestCase):
                 self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 self.assertManholeRunning(proc, uds_path, oneshot=True)
 
+    def test_fail_to_cry(self):
+        import manhole
+        verbose = manhole.VERBOSE
+        out = manhole._STDERR
+        try:
+            manhole.VERBOSE = True
+            fh = os.fdopen(os.dup(2), 'w')
+            fh.close()
+            manhole._STDERR = fh
+            manhole.cry('stuff')
+        finally:
+            manhole.VERBOSE = verbose
+            manhole._STDERR = out
+
+    def test_oneshot_on_usr2_error(self):
+        with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_oneshot_on_usr2') as proc:
+            with self._dump_on_error(proc.read):
+                self._wait_for_strings(proc.read, 10, 'Not patching os.fork and os.forkpty. Oneshot activation is done by signal 12')
+                self.assertRaises(AssertionError, self._wait_for_strings, proc.read, 2, '/tmp/manhole-')
+                proc.signal(signal.SIGUSR2)
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
+                uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                self.assertManholeRunning(proc, uds_path, oneshot=True, extra=lambda sock: sock.send(b"raise SystemExit()\n"))
+
+                proc.signal(signal.SIGUSR2)
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
+                uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                self.assertManholeRunning(proc, uds_path, oneshot=True)
+
+    def test_interrupt_on_accept(self):
+        with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_interrupt_on_accept') as proc:
+            with self._dump_on_error(proc.read):
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
+                uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection', 'Sending signal to manhole thread', 'Waiting for new connection')
+                self.assertManholeRunning(proc, uds_path)
+
 cov = None
 def maybe_enable_coverage():
     global cov
+    try:
+        from coverage.control import coverage
+        from coverage.collector import Collector
+    except ImportError:
+        cov = None
+        return
     if cov:
+        cov.save()
         cov.stop()
+    if Collector._collectors:
+        Collector._collectors[-1].stop()
     cov = cov or os.environ.get("WITH_COVERAGE")
     if cov:
-        try:
-            from coverage.control import coverage
-        except ImportError:
-            cov = None
-            return
         cov = coverage(auto_data=True, data_suffix=True, timid=False, include=['src/*'])
         cov.start()
 
         @atexit.register
         def cleanup():
-            cov.stop()
+            if cov.collector._collectors:
+                cov.stop()
             cov.save()
 
-
 def monkeypatch(mod, what):
     old = getattr(mod, what)
     def decorator(func):
@@ -426,6 +471,29 @@ if __name__ == '__main__':
             manhole.install(activate_on='USR2', oneshot_on='USR2')
             for i in range(100):
                 time.sleep(0.1)
+        elif test_name == 'test_interrupt_on_accept':
+            def handle_usr2(_sig, _frame):
+                print('Got USR2')
+            signal.signal(signal.SIGUSR2, handle_usr2)
+
+            import ctypes
+            import ctypes.util
+            libpthread_path = ctypes.util.find_library("pthread")
+            if not libpthread_path:
+                raise ImportError
+            libpthread = ctypes.CDLL(libpthread_path)
+            if not hasattr(libpthread, "pthread_setname_np"):
+                raise ImportError
+            pthread_kill = libpthread.pthread_kill
+            pthread_kill.argtypes = [ctypes.c_void_p, ctypes.c_int]
+            pthread_kill.restype = ctypes.c_int
+            manhole.install(sigmask=None)
+            for i in range(15):
+                time.sleep(0.1)
+            print("Sending signal to manhole thread ...")
+            pthread_kill(manhole._INST.ident, signal.SIGUSR2)
+            for i in range(100):
+                time.sleep(0.1)
         elif test_name == 'test_oneshot_on_usr2':
             manhole.install(oneshot_on='USR2')
             for i in range(100):
